# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-12-27 11:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0027_auto_20201123_1552"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pipelinetemplate",
            name="create_time",
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name="创建时间"),
        ),
        migrations.AlterField(
            model_name="pipelinetemplate",
            name="edit_time",
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name="修改时间"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import contextlib

from pipeline.celery.settings import QueueResolver
from pipeline.engine import tasks, exceptions
from pipeline.engine.models import (
    NodeCeleryTask,
    PipelineModel,
    PipelineProcess,
    ProcessCeleryTask,
    ScheduleCeleryTask,
    SendFailedCeleryTask,
)

logger = logging.getLogger("root")


@contextlib.contextmanager
def celery_task_send_fail_pass():
    try:
        yield
    except exceptions.CeleryFailedTaskCatchException as e:
        # we catch CeleryFailedTaskCatchException here and ignore it.
        # so we can process the fail task in SendFailedCeleryTask
        logger.exception("{} task send error.".format(e.task_name))


class CeleryTaskArgsResolver(object):
    def __init__(self, process_id):
        self.process_id = process_id

    def resolve_args(self, task):
        args = {}
        task_args = PipelineProcess.objects.task_args_for_process(self.process_id)

        queue = task_args["queue"]
        priority = task_args["priority"]

        args["priority"] = priority

        if queue:
            args["routing_key"] = QueueResolver(queue).resolve_task_routing_key(task)

        return args


def pipeline_ready_handler(sender, process_id, **kwargs):
    task = tasks.start
    args_resolver = CeleryTaskArgsResolver(process_id)

    with celery_task_send_fail_pass():
        ProcessCeleryTask.objects.start_task(
            process_id=process_id, task=task, kwargs={"args": [process_id], **args_resolver.resolve_args(task)},
        )


def pipeline_end_handler(sender, root_pipeline_id, **kwargs):
    pass


def child_process_ready_handler(sender, child_id, **kwargs):
    task = tasks.dispatch
    args_resolver = CeleryTaskArgsResolver(child_id)

    with celery_task_send_fail_pass():
        ProcessCeleryTask.objects.start_task(
            process_id=child_id, task=task, kwargs={"args": [child_id], **args_resolver.resolve_args(task)},
        )


def process_ready_handler(sender, process_id, current_node_id=None, call_from_child=False, **kwargs):

    task = tasks.process_wake_up
    args_resolver = CeleryTaskArgsResolver(process_id)

    with celery_task_send_fail_pass():
        ProcessCeleryTask.objects.start_task(
            process_id=process_id,
            task=task,
            kwargs={"args": [process_id, current_node_id, call_from_child], **args_resolver.resolve_args(task)},
        )


def batch_process_ready_handler(sender, process_id_list, pipeline_id, **kwargs):

    task = tasks.batch_wake_up
    task_args = PipelineModel.objects.task_args_for_pipeline(pipeline_id)
    priority = task_args["priority"]
    queue = task_args["queue"]

    kwargs = {
        "args": [process_id_list, pipeline_id],
        "priority": priority,
    }
    if queue:
        kwargs["routing_key"] = QueueResolver(queue).resolve_task_routing_key(task)

    with celery_task_send_fail_pass():
        with SendFailedCeleryTask.watch(
            name=task.name, kwargs=kwargs, type=SendFailedCeleryTask.TASK_TYPE_EMPTY, extra_kwargs={},
        ):
            task.apply_async(**kwargs)


def wake_from_schedule_handler(sender, process_id, activity_id, **kwargs):

    task = tasks.wake_from_schedule
    args_resolver = CeleryTaskArgsResolver(process_id)

    with celery_task_send_fail_pass():
        ProcessCeleryTask.objects.start_task(
            process_id=process_id,
            task=task,
            kwargs={"args": [process_id, activity_id], **args_resolver.resolve_args(task)},
        )


def process_unfreeze_handler(sender, process_id, **kwargs):
    task = tasks.process_unfreeze
    args_resolver = CeleryTaskArgsResolver(process_id)

    with celery_task_send_fail_pass():
        ProcessCeleryTask.objects.start_task(
            process_id=process_id, task=task, kwargs={"args": [process_id], **args_resolver.resolve_args(task)},
        )


def schedule_ready_handler(sender, process_id, schedule_id, countdown, data_id=None, **kwargs):
    task = tasks.service_schedule
    args_resolver = CeleryTaskArgsResolver(process_id)

    with celery_task_send_fail_pass():
        ScheduleCeleryTask.objects.start_task(
            schedule_id=schedule_id,
            task=task,
            kwargs={
                "args": [process_id, schedule_id, data_id],
                "countdown": countdown,
                **args_resolver.resolve_args(task),
            },
        )


def service_activity_timeout_monitor_start_handler(sender, node_id, version, root_pipeline_id, countdown, **kwargs):
    NodeCeleryTask.objects.start_task(
        node_id=node_id,
        task=tasks.node_timeout_check,
        kwargs={
            "args": [node_id, version, root_pipeline_id],
            "countdown": countdown,
            "priority": PipelineModel.objects.priority_for_pipeline(root_pipeline_id),
        },
    )


def service_activity_timeout_monitor_end_handler(sender, node_id, version, **kwargs):
    NodeCeleryTask.objects.revoke(node_id)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2019-09-12 07:31


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0019_auto_20190729_1044"),
    ]

    operations = [
        migrations.AddField(
            model_name="pipelinemodel",
            name="priority",
            field=models.IntegerField(default=100, verbose_name="\u6d41\u7a0b\u4f18\u5148\u7ea7"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json
from typing import Dict

from bamboo_engine import metrics
from bamboo_engine.eri import Data, DataInput, ExecutionData, CallbackData

from pipeline.eri.models import Data as DBData
from pipeline.eri.models import ExecutionData as DBExecutionData
from pipeline.eri.models import CallbackData as DBCallbackData
from pipeline.eri.imp.serializer import SerializerMixin


class DataMixin(SerializerMixin):
    def _get_data_inputs(self, inputs: dict):
        return {k: DataInput(need_render=v["need_render"], value=v["value"]) for k, v in inputs.items()}

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_DATA_READ_TIME)
    def get_data(self, node_id: str) -> Data:
        """
        获取某个节点的数据对象

        :param node_id: 节点 ID
        :type node_id: str
        :return: 数据对象实例
        :rtype: Data
        """

        data_model = DBData.objects.get(node_id=node_id)
        return Data(inputs=self._get_data_inputs(json.loads(data_model.inputs)), outputs=json.loads(data_model.outputs))

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_DATA_INPUTS_READ_TIME)
    def get_data_inputs(self, node_id: str) -> Dict[str, DataInput]:
        """
        获取某个节点的输入数据

        :param node_id: 节点 ID
        :type node_id: str
        :return: 输入数据字典
        :rtype: dict
        """
        qs = DBData.objects.filter(node_id=node_id).only("inputs")

        if not qs:
            raise DBData.DoesNotExist()

        return self._get_data_inputs(json.loads(qs[0].inputs))

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_DATA_OUTPUTS_READ_TIME)
    def get_data_outputs(self, node_id: str) -> dict:
        """
        获取某个节点的输出数据

        :param node_id: 节点 ID
        :type node_id: str
        :return: 输入数据字典
        :rtype: dict
        """
        qs = DBData.objects.filter(node_id=node_id).only("outputs")

        if not qs:
            raise DBData.DoesNotExist()

        return json.loads(qs[0].outputs)

    def set_data_inputs(self, node_id: str, data: Dict[str, DataInput]):
        """
        将节点数据对象的 inputs 设置为 data

        : param node_id: 节点 ID
        : type node_id: str
        : param data: 目标数据
        : type data: dict
        """
        inputs = json.dumps({k: {"need_render": v.need_render, "value": v.value} for k, v in data.items()})
        if DBData.objects.filter(node_id=node_id).exists():
            DBData.objects.filter(node_id=node_id).update(inputs=inputs)
        else:
            DBData.objects.create(node_id=node_id, inputs=inputs, outputs="{}")

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_READ_TIME)
    def get_execution_data(self, node_id: str) -> ExecutionData:
        """
        获取某个节点的执行数据

        : param node_id: 节点 ID
        : type node_id: str
        : return: 执行数据实例
        : rtype: ExecutionData
        """
        data_model = DBExecutionData.objects.get(node_id=node_id)
        return ExecutionData(
            inputs=self._deserialize(data_model.inputs, data_model.inputs_serializer),
            outputs=self._deserialize(data_model.outputs, data_model.outputs_serializer),
        )

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_INPUTS_READ_TIME)
    def get_execution_data_inputs(self, node_id: str) -> dict:
        """
        获取某个节点的执行数据输入

        :param node_id: 节点 ID
        :type node_id: str
        :return: 执行数据输入
        :rtype: dict
        """
        qs = DBExecutionData.objects.filter(node_id=node_id).only("inputs_serializer", "inputs")

        if not qs:
            return {}

        return self._deserialize(qs[0].inputs, qs[0].inputs_serializer)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_OUTPUTS_READ_TIME)
    def get_execution_data_outputs(self, node_id: str) -> dict:
        """
        获取某个节点的执行数据输出

        :param node_id: 节点 ID
        :type node_id: str
        :return: 执行数据输出
        :rtype: dict
        """
        qs = DBExecutionData.objects.filter(node_id=node_id).only("outputs_serializer", "outputs")

        if not qs:
            return {}

        return self._deserialize(qs[0].outputs, qs[0].outputs_serializer)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_WRITE_TIME)
    def set_execution_data(self, node_id: str, data: ExecutionData):
        """
        设置某个节点的执行数据

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 执行数据实例
        :type data: ExecutionData
        """
        inputs, inputs_serializer = self._serialize(data.inputs)
        outputs, outputs_serializer = self._serialize(data.outputs)
        if DBExecutionData.objects.filter(node_id=node_id).exists():
            DBExecutionData.objects.filter(node_id=node_id).update(
                inputs=inputs,
                inputs_serializer=inputs_serializer,
                outputs=outputs,
                outputs_serializer=outputs_serializer,
            )
        else:
            DBExecutionData.objects.create(
                node_id=node_id,
                inputs=inputs,
                inputs_serializer=inputs_serializer,
                outputs=outputs,
                outputs_serializer=outputs_serializer,
            )

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_INPUTS_WRITE_TIME)
    def set_execution_data_inputs(self, node_id: str, inputs: dict):
        """
        设置某个节点的执行数据输入

        :param node_id: 节点 ID
        :type node_id: str
        :param outputs: 输出数据
        :type outputs: dict
        """
        inputs, inputs_serializer = self._serialize(inputs)
        if DBExecutionData.objects.filter(node_id=node_id).exists():
            DBExecutionData.objects.filter(node_id=node_id).update(inputs=inputs, inputs_serializer=inputs_serializer)
        else:
            DBExecutionData.objects.create(
                node_id=node_id,
                inputs=inputs,
                inputs_serializer=inputs_serializer,
                outputs="{}",
                outputs_serializer=self.JSON_SERIALIZER,
            )

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_EXEC_DATA_OUTPUTS_WRITE_TIME)
    def set_execution_data_outputs(self, node_id: str, outputs: dict):
        """
        设置某个节点的执行数据输出

        :param node_id: 节点 ID
        :type node_id: str
        :param outputs: 输出数据
        :type outputs: dict
        """
        outputs, outputs_serializer = self._serialize(outputs)
        if DBExecutionData.objects.filter(node_id=node_id).exists():
            DBExecutionData.objects.filter(node_id=node_id).update(
                outputs=outputs, outputs_serializer=outputs_serializer
            )
        else:
            DBExecutionData.objects.create(
                node_id=node_id,
                inputs="{}",
                inputs_serializer=self.JSON_SERIALIZER,
                outputs=outputs,
                outputs_serializer=outputs_serializer,
            )

    def set_callback_data(self, node_id: str, version: str, data: dict) -> int:
        """
        设置某个节点执行数据的回调数据

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: dict
        :return: 回调数据 ID
        :rtype: int
        """
        return DBCallbackData.objects.create(node_id=node_id, version=version, data=json.dumps(data)).id

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_CALLBACK_DATA_READ_TIME)
    def get_callback_data(self, data_id: int) -> CallbackData:
        """
        获取回调数据

        :param data_id: Data ID
        :type data_id: int
        :return: 回调数据实例
        :rtype: CallbackData
        """
        data_model = DBCallbackData.objects.get(id=data_id)
        return CallbackData(
            id=data_model.id, node_id=data_model.node_id, version=data_model.version, data=json.loads(data_model.data)
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import logging
import ujson as json
from copy import deepcopy

from celery import task
from bamboo_engine import api as bamboo_engine_api

from pipeline.component_framework.constants import LEGACY_PLUGINS_VERSION
from pipeline.contrib.statistics.models import (
    ComponentExecuteData,
    InstanceInPipeline,
)
from pipeline.contrib.statistics.utils import count_pipeline_tree_nodes
from pipeline.core.constants import PE
from pipeline.engine import api as pipeline_api
from pipeline.engine import states
from pipeline.engine.exceptions import InvalidOperationException
from pipeline.engine.utils import calculate_elapsed_time
from pipeline.models import PipelineInstance
from pipeline.eri.runtime import BambooDjangoRuntime

logger = logging.getLogger("celery")


def recursive_collect_components(activities, status_tree, instance_id, stack=None, engine_ver=1):
    """
    @summary 递归流程树，获取所有执行成功/失败的插件
    @param activities: 当前流程树的任务节点信息
    @param status_tree: 当前流程树的任务节点状态
    @param instance_id: 根流程的示例 instance_id
    @param stack: 子流程堆栈
    """
    if stack is None:
        stack = []
        is_sub = False
    else:
        is_sub = True
    component_list = []
    for act_id, act in activities.items():
        # 只有执行了才会查询到 status，兼容中途撤销的任务
        if act_id in status_tree:
            exec_act = status_tree[act_id]
            # 属于标准插件节点
            if act[PE.type] == PE.ServiceActivity:
                if exec_act["state"] in states.ARCHIVED_STATES:
                    create_kwargs = {
                        "component_code": act["component"]["code"],
                        "instance_id": instance_id,
                        "is_sub": is_sub,
                        "node_id": act_id,
                        "subprocess_stack": json.dumps(stack),
                        "started_time": exec_act["started_time"],
                        "archived_time": exec_act["archived_time"],
                        "elapsed_time": exec_act.get(
                            "elapsed_time", calculate_elapsed_time(exec_act["started_time"], exec_act["archived_time"])
                        ),
                        "is_skip": exec_act["skip"],
                        "is_retry": False,
                        "status": exec_act["state"] == "FINISHED",
                        "version": act["component"].get("version", LEGACY_PLUGINS_VERSION),
                    }
                    component_list.append(ComponentExecuteData(**create_kwargs))
                    if exec_act["retry"] > 0:
                        # 需要通过执行历史获得
                        if engine_ver == 1:
                            history_list = pipeline_api.get_activity_histories(act_id)
                        else:
                            history_list = bamboo_engine_api.get_node_short_histories(
                                runtime=BambooDjangoRuntime(), node_id=act_id
                            )

                        for history in history_list:
                            create_kwargs.update(
                                {
                                    "started_time": history["started_time"],
                                    "archived_time": history["archived_time"],
                                    "elapsed_time": history.get(
                                        "elapsed_time",
                                        calculate_elapsed_time(history["started_time"], history["archived_time"]),
                                    ),
                                    "is_retry": True,
                                    "is_skip": False,
                                    "status": False,
                                }
                            )
                            component_list.append(ComponentExecuteData(**create_kwargs))
            # 子流程的执行堆栈（子流程的执行过程）
            elif act[PE.type] == PE.SubProcess:
                # 递归子流程树
                sub_activities = act[PE.pipeline][PE.activities]
                # 防止stack共用
                copied_stack = deepcopy(stack)
                copied_stack.insert(0, act_id)
                component_list += recursive_collect_components(
                    sub_activities, exec_act["children"], instance_id, copied_stack
                )
    return component_list


@task
def pipeline_post_save_statistics_task(instance_id):
    instance = PipelineInstance.objects.get(instance_id=instance_id)
    # 统计流程标准插件个数，子流程个数，网关个数
    try:
        atom_total, subprocess_total, gateways_total = count_pipeline_tree_nodes(instance.execution_data)
        InstanceInPipeline.objects.update_or_create(
            instance_id=instance_id,
            defaults={
                "atom_total": atom_total,
                "subprocess_total": subprocess_total,
                "gateways_total": gateways_total,
            },
        )
    except Exception as e:
        logger.error(
            (
                "pipeline_post_save_handler save InstanceInPipeline[instance_id={instance_id}] " "raise error: {error}"
            ).format(instance_id=instance_id, error=e)
        )


@task
def pipeline_archive_statistics_task(instance_id):
    instance = PipelineInstance.objects.get(instance_id=instance_id)
    engine_ver = 1
    # 获得任务实例的执行树
    try:
        status_tree = pipeline_api.get_status_tree(instance_id, 99)
    except InvalidOperationException:
        engine_ver = 2
        status_tree_result = bamboo_engine_api.get_pipeline_states(
            runtime=BambooDjangoRuntime(), root_id=instance_id, flat_children=False
        )
        if not status_tree_result.result:
            logger.error(
                "pipeline_archive_statistics_task bamboo_engine_api.get_pipeline_states fail: {}".format(
                    status_tree_result.result.exc_trace
                )
            )
            return
        status_tree = status_tree_result.data[instance_id]

    # 删除原有标准插件数据
    ComponentExecuteData.objects.filter(instance_id=instance_id).delete()
    # 获得任务实例的执行数据
    data = instance.execution_data
    try:
        component_list = recursive_collect_components(
            activities=data[PE.activities],
            status_tree=status_tree["children"],
            instance_id=instance_id,
            engine_ver=engine_ver,
        )
        ComponentExecuteData.objects.bulk_create(component_list)
    except Exception:
        logger.exception(
            ("pipeline_post_save_handler save ComponentExecuteData[instance_id={instance_id}] raise error").format(
                instance_id=instance_id
            )
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-13 07:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0022_scheduleservice_multi_callback_enabled"),
    ]

    operations = [
        migrations.AddField(
            model_name="status",
            name="state_refresh_at",
            field=models.DateTimeField(null=True, verbose_name="上次状态更新的时间"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import functools
import logging
import time

from celery import current_app
from django.db import transaction
from redis.exceptions import ConnectionError as RedisConnectionError

from pipeline.celery.queues import ScalableQueues
from pipeline.constants import PIPELINE_DEFAULT_PRIORITY, PIPELINE_MAX_PRIORITY, PIPELINE_MIN_PRIORITY
from pipeline.core.flow.activity import ServiceActivity
from pipeline.core.flow.gateway import ExclusiveGateway, ParallelGateway
from pipeline.engine import exceptions, states
from pipeline.engine.core.api import workers
from pipeline.engine.models import (
    Data,
    FunctionSwitch,
    History,
    NodeRelationship,
    Pipeline,
    PipelineModel,
    PipelineProcess,
    ProcessCeleryTask,
    ScheduleService,
    Status,
    SubProcessRelationship,
)
from pipeline.engine.signals import pipeline_revoke
from pipeline.engine.utils import ActionResult, calculate_elapsed_time
from pipeline.utils import uniqid

logger = logging.getLogger("celery")


def _node_existence_check(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        id_from_kwargs = kwargs.get("node_id")
        node_id = id_from_kwargs if id_from_kwargs else args[0]
        try:
            Status.objects.get(id=node_id)
        except Status.DoesNotExist:
            return ActionResult(result=False, message="node not exists or not be executed yet")
        return func(*args, **kwargs)

    return wrapper


def _frozen_check(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if FunctionSwitch.objects.is_frozen():
            return ActionResult(result=False, message="engine is frozen, can not perform operation")

        return func(*args, **kwargs)

    return wrapper


def _worker_check(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        def on_connection_error(exc, interval):
            logger.warning("Connection Error: {!r}. Retry in {}s.".format(exc, interval))

        if kwargs.get("check_workers", True):
            try:
                with current_app.connection() as conn:
                    try:
                        conn.ensure_connection(on_connection_error, current_app.conf.BROKER_CONNECTION_MAX_RETRIES)
                    except conn.connection_errors + conn.channel_errors as exc:
                        logger.warning("Connection lost: {!r}".format(exc))
                    if not workers(conn):
                        return ActionResult(
                            result=False, message="can not find celery workers, please check worker status"
                        )
            except exceptions.RabbitMQConnectionError as e:
                return ActionResult(
                    result=False,
                    message="celery worker status check failed with message: %s, " "check rabbitmq status please" % e,
                )
            except RedisConnectionError:
                return ActionResult(result=False, message="redis connection error, check redis status please")

        return func(*args, **kwargs)

    return wrapper


@_worker_check
@_frozen_check
def start_pipeline(pipeline_instance, check_workers=True, priority=PIPELINE_DEFAULT_PRIORITY, queue=""):
    """
    start a pipeline
    :param pipeline_instance:
    :param priority:
    :return:
    """

    if priority > PIPELINE_MAX_PRIORITY or priority < PIPELINE_MIN_PRIORITY:
        raise exceptions.InvalidOperationException(
            "pipeline priority must between [{min}, {max}]".format(min=PIPELINE_MIN_PRIORITY, max=PIPELINE_MAX_PRIORITY)
        )

    if queue and not ScalableQueues.has_queue(queue):
        return ActionResult(result=False, message="can't not find queue({}) in any config queues.".format(queue))

    Status.objects.prepare_for_pipeline(pipeline_instance)
    process = PipelineProcess.objects.prepare_for_pipeline(pipeline_instance)
    PipelineModel.objects.prepare_for_pipeline(pipeline_instance, process, priority, queue=queue)

    PipelineModel.objects.pipeline_ready(process_id=process.id)

    return ActionResult(result=True, message="success")


@_frozen_check
def pause_pipeline(pipeline_id):
    """
    pause a running pipeline
    :param pipeline_id:
    :return:
    """

    return Status.objects.transit(id=pipeline_id, to_state=states.SUSPENDED, is_pipeline=True, appoint=True)


@_worker_check
@_frozen_check
def resume_pipeline(pipeline_id):
    """
    resume a pipeline from suspended
    :param pipeline_id:
    :return:
    """

    action_result = Status.objects.transit(id=pipeline_id, to_state=states.READY, is_pipeline=True, appoint=True)
    if not action_result.result:
        return action_result

    process = PipelineModel.objects.get(id=pipeline_id).process
    to_be_waked = []
    _get_process_to_be_waked(process, to_be_waked)
    PipelineProcess.objects.batch_process_ready(process_id_list=to_be_waked, pipeline_id=pipeline_id)

    return action_result


@_frozen_check
def revoke_pipeline(pipeline_id):
    """
    revoke a pipeline
    :param pipeline_id:
    :return:
    """

    try:
        pipeline_model = PipelineModel.objects.get(id=pipeline_id)
    except PipelineModel.DoesNotExist:
        return ActionResult(result=False, message="pipeline to be revoked does not exist.")

    action_result = Status.objects.transit(id=pipeline_id, to_state=states.REVOKED, is_pipeline=True, appoint=True)
    if not action_result.result:
        return action_result

    process = pipeline_model.process

    if not process:
        return ActionResult(result=False, message="relate process is none, this pipeline may be revoked.")

    with transaction.atomic():
        PipelineProcess.objects.select_for_update().get(id=process.id)
        process.revoke_subprocess()
        process.destroy_all()

    pipeline_revoke.send(sender=Pipeline, root_pipeline_id=pipeline_id)

    return action_result


@_frozen_check
def pause_node_appointment(node_id):
    """
    make a appointment to pause a node
    :param node_id:
    :return:
    """

    return Status.objects.transit(id=node_id, to_state=states.SUSPENDED, appoint=True)


@_worker_check
@_frozen_check
@_node_existence_check
def resume_node_appointment(node_id):
    """
    make a appointment to resume a node
    :param node_id:
    :return:
    """

    qs = PipelineProcess.objects.filter(current_node_id=node_id, is_sleep=True)
    if qs.exists():
        # a process had sleep caused by pause reservation
        action_result = Status.objects.transit(id=node_id, to_state=states.READY, appoint=True)
        if not action_result.result:
            return action_result

        process = qs.first()
        Status.objects.recover_from_block(process.root_pipeline.id, process.subprocess_stack)
        PipelineProcess.objects.process_ready(process_id=process.id)
        return ActionResult(result=True, message="success")

    processing_sleep = SubProcessRelationship.objects.get_relate_process(subprocess_id=node_id)
    if processing_sleep.exists():
        action_result = Status.objects.transit(id=node_id, to_state=states.RUNNING, appoint=True, is_pipeline=True)
        if not action_result.result:
            return action_result
        # processes had sleep caused by subprocess pause
        root_pipeline_id = processing_sleep.first().root_pipeline_id

        process_can_be_waked = [p for p in processing_sleep if p.can_be_waked()]
        can_be_waked_ids = [p.id for p in process_can_be_waked]

        # get subprocess id which should be transited
        subprocess_to_be_transit = set()
        for process in process_can_be_waked:
            _, subproc_above = process.subproc_sleep_check()
            for subproc in subproc_above:
                subprocess_to_be_transit.add(subproc)

        Status.objects.recover_from_block(root_pipeline_id, subprocess_to_be_transit)
        PipelineProcess.objects.batch_process_ready(process_id_list=can_be_waked_ids, pipeline_id=root_pipeline_id)
        return ActionResult(result=True, message="success")

    return ActionResult(result=False, message="node not exists or not be executed yet")


@_worker_check
@_frozen_check
@_node_existence_check
def retry_node(node_id, inputs=None):
    """
    retry a node
    :param node_id:
    :param inputs:
    :return:
    """

    try:
        PipelineProcess.objects.get(current_node_id=node_id)
    except PipelineProcess.DoesNotExist:  # can not retry subprocess
        return ActionResult(result=False, message="can't not retry a subprocess or this process has been revoked")

    process = PipelineProcess.objects.get(current_node_id=node_id)

    # try to get next
    node = process.top_pipeline.node(node_id)
    if not (isinstance(node, ServiceActivity) or isinstance(node, ParallelGateway)):
        return ActionResult(result=False, message="can't retry this type of node")

    if hasattr(node, "retryable") and not node.retryable:
        return ActionResult(result=False, message="the node is set to not be retryable, try skip it please.")

    action_result = Status.objects.retry(process, node, inputs)
    if not action_result.result:
        return action_result

    # wake up process
    PipelineProcess.objects.process_ready(process_id=process.id)

    return action_result


@_worker_check
@_frozen_check
@_node_existence_check
def skip_node(node_id):
    """
    skip a node
    :param node_id:
    :return:
    """

    try:
        process = PipelineProcess.objects.get(current_node_id=node_id)
    except PipelineProcess.DoesNotExist:  # can not skip subprocess
        return ActionResult(result=False, message="can't not skip a subprocess or this process has been revoked")

    # try to get next
    node = process.top_pipeline.node(node_id)
    if not isinstance(node, ServiceActivity):
        return ActionResult(result=False, message="can't skip this type of node")

    if hasattr(node, "skippable") and not node.skippable:
        return ActionResult(result=False, message="this node is set to not be skippable, try retry it please.")

    # skip and write result bit
    action_result = Status.objects.skip(process, node)
    if not action_result.result:
        return action_result

    next_node_id = node.next().id

    # extract outputs and wake up process
    process.top_pipeline.context.extract_output(node)
    process.save()
    PipelineProcess.objects.process_ready(process_id=process.id, current_node_id=next_node_id)

    return action_result


@_worker_check
@_frozen_check
@_node_existence_check
def skip_exclusive_gateway(node_id, flow_id):
    """
    skip a failed exclusive gateway and appoint the flow to be pushed
    :param node_id:
    :param flow_id:
    :return:
    """

    try:
        process = PipelineProcess.objects.get(current_node_id=node_id)
    except PipelineProcess.DoesNotExist:
        return ActionResult(
            result=False, message="invalid operation, this gateway is finished or pipeline have been revoked"
        )

    exclusive_gateway = process.top_pipeline.node(node_id)

    if not isinstance(exclusive_gateway, ExclusiveGateway):
        return ActionResult(result=False, message="invalid operation, this node is not a exclusive gateway")

    next_node_id = exclusive_gateway.target_for_sequence_flow(flow_id).id

    action_result = Status.objects.skip(process, exclusive_gateway)
    if not action_result.result:
        return action_result

    # wake up process
    PipelineProcess.objects.process_ready(process_id=process.id, current_node_id=next_node_id)

    return action_result


def get_status_tree(node_id, max_depth=1):
    """
    get state and children states for a node
    :param node_id:
    :param max_depth:
    :return:
    """
    rel_qs = NodeRelationship.objects.filter(ancestor_id=node_id, distance__lte=max_depth)
    if not rel_qs.exists():
        raise exceptions.InvalidOperationException("node(%s) does not exist, may have not by executed" % node_id)
    descendants = [rel.descendant_id for rel in rel_qs]
    # remove root node
    descendants.remove(node_id)

    rel_qs = NodeRelationship.objects.filter(descendant_id__in=descendants, distance=1)
    targets = [rel.descendant_id for rel in rel_qs]

    root_status = Status.objects.filter(id=node_id).values().first()
    root_status["elapsed_time"] = calculate_elapsed_time(root_status["started_time"], root_status["archived_time"])
    status_map = {node_id: root_status}
    status_qs = Status.objects.filter(id__in=targets).values()
    for status in status_qs:
        status["elapsed_time"] = calculate_elapsed_time(status["started_time"], status["archived_time"])
        status_map[status["id"]] = status

    relationships = [(s.ancestor_id, s.descendant_id) for s in rel_qs]
    for (parent_id, child_id) in relationships:
        if parent_id not in status_map:
            return

        parent_status = status_map[parent_id]
        child_status = status_map[child_id]
        child_status.setdefault("children", {})

        parent_status.setdefault("children", {}).setdefault(child_id, child_status)

    return status_map[node_id]


@_worker_check
@_frozen_check
def activity_callback(activity_id, callback_data):
    """
    callback a schedule node
    :param activity_id:
    :param callback_data:
    :return:
    """

    version = Status.objects.version_for(activity_id)
    times = 0

    # it's possible that ScheduleService is not be set when callback make
    while times < 3:
        try:
            service = ScheduleService.objects.schedule_for(activity_id, version)
            break
        except ScheduleService.DoesNotExist as e:
            times += 1
            time.sleep(times)
            if times >= 3:
                raise e

    try:
        process_id = PipelineProcess.objects.get(current_node_id=activity_id).id
    except PipelineProcess.DoesNotExist:
        return ActionResult(
            result=False, message="invalid operation, this node is finished or pipeline have been revoked"
        )

    if service.is_finished:
        raise exceptions.InvalidOperationException("activity(%s) callback already finished" % activity_id)
    service.callback(callback_data, process_id)
    return ActionResult(result=True, message="success")


def get_inputs(node_id):
    """
    get inputs data for a node
    :param node_id:
    :return:
    """
    return Data.objects.get(id=node_id).inputs


def get_outputs(node_id):
    """
    get outputs data for a node
    :param node_id:
    :return:
    """
    data = Data.objects.get(id=node_id)
    return {"outputs": data.outputs, "ex_data": data.ex_data}


def get_batch_outputs(node_ids):
    """
    get outputs data for a batch of nodes
    :param node_ids: a list of node_id
    :return:
    """
    nodes_data = Data.objects.filter(id__in=node_ids)
    return {node_data.id: {"outputs": node_data.outputs, "ex_data": node_data.ex_data} for node_data in nodes_data}


def get_activity_histories(node_id, loop=None):
    """
    get get_activity_histories data for a node
    :param node_id: 节点 ID
    :param loop: 循环序号
    :return:
    """
    return History.objects.get_histories(node_id, loop)


@_frozen_check
@_node_existence_check
def forced_fail(node_id, kill=False, ex_data=""):
    """
    forced fail a node
    :param node_id:
    :param kill:
    :param ex_data:
    :return:
    """

    try:
        process = PipelineProcess.objects.get(current_node_id=node_id)
    except PipelineProcess.DoesNotExist:
        return ActionResult(
            result=False, message="invalid operation, this node is finished or pipeline have been revoked"
        )

    node = process.top_pipeline.node(node_id)
    if not isinstance(node, ServiceActivity):
        return ActionResult(result=False, message="can't not forced fail this type of node")

    action_result = Status.objects.transit(node_id, to_state=states.FAILED)
    if not action_result.result:
        return action_result

    try:
        node.failure_handler(process.root_pipeline.data)
    except Exception:
        pass

    with transaction.atomic():
        s = Status.objects.get(id=node.id)
        ScheduleService.objects.delete_schedule(s.id, s.version)
        Data.objects.forced_fail(node_id, ex_data)
        ProcessCeleryTask.objects.revoke(process.id, kill)
        process.adjust_status()
        process.is_sleep = True
        process.save()
        s.version = uniqid.uniqid()
        s.save()

    return ActionResult(result=True, message="success")


def _get_process_to_be_waked(process, to_be_waked):
    if process.can_be_waked():
        to_be_waked.append(process.id)
    elif process.children:
        for child_id in process.children:
            child = PipelineProcess.objects.get(id=child_id)
            _get_process_to_be_waked(child, to_be_waked)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from pipeline.core.flow.gateway import ConvergeGateway
from pipeline.engine import exceptions
from pipeline.engine.models import Status

from .base import FlowElementHandler

logger = logging.getLogger("celery")

__all__ = ["ConvergeGatewayHandler"]


class ConvergeGatewayHandler(FlowElementHandler):
    @staticmethod
    def element_cls():
        return ConvergeGateway

    def handle(self, process, element, status):
        # try to sync data if current process has children
        if process.children:
            try:
                process.sync_with_children()
            except exceptions.ChildDataSyncError:
                logger.error(traceback.format_exc())
                # clean children and update current_node to prevent re execute child process
                process.clean_children()
                Status.objects.fail(element, ex_data="Sync branch context error, check data backend status please.")
                return self.HandleResult(next_node=None, should_return=True, should_sleep=True)

        Status.objects.finish(element)
        return self.HandleResult(next_node=element.next(), should_return=False, should_sleep=False)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import transaction
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

from pipeline.core.constants import PE
from pipeline.core.pipeline import Pipeline
from pipeline.engine.signals import pipeline_end, pipeline_revoke
from pipeline.models import (
    PipelineInstance,
    PipelineTemplate,
    TemplateCurrentVersion,
    TemplateRelationship,
    TemplateVersion,
)


@receiver(pre_save, sender=PipelineTemplate)
def pipeline_template_pre_save_handler(sender, instance, **kwargs):
    template = instance

    if template.is_deleted:
        return

    template.set_has_subprocess_bit()


@receiver(post_save, sender=PipelineTemplate)
def pipeline_template_post_save_handler(sender, instance, created, **kwargs):
    template = instance

    if template.is_deleted:
        TemplateRelationship.objects.filter(ancestor_template_id=template.template_id).delete()
        return

    with transaction.atomic():
        TemplateRelationship.objects.filter(ancestor_template_id=template.template_id).delete()
        acts = list(template.data[PE.activities].values())
        subprocess_nodes = [act for act in acts if act["type"] == PE.SubProcess]
        rs = []
        for sp in subprocess_nodes:
            version = sp.get("version") or PipelineTemplate.objects.get(template_id=sp["template_id"]).version
            rs.append(
                TemplateRelationship(
                    ancestor_template_id=template.template_id,
                    descendant_template_id=sp["template_id"],
                    subprocess_node_id=sp["id"],
                    version=version,
                )
            )
        if rs:
            TemplateRelationship.objects.bulk_create(rs)
        TemplateVersion.objects.track(template)
        TemplateCurrentVersion.objects.update_current_version(template)


@receiver(pipeline_end, sender=Pipeline)
def pipeline_end_handler(sender, root_pipeline_id, **kwargs):
    try:
        PipelineInstance.objects.set_finished(root_pipeline_id)
    except PipelineInstance.DoesNotExist:  # task which do not belong to any instance
        pass


@receiver(pipeline_revoke, sender=Pipeline)
def pipeline_revoke_handler(sender, root_pipeline_id, **kwargs):
    try:
        PipelineInstance.objects.set_revoked(root_pipeline_id)
    except PipelineInstance.DoesNotExist:  # task which do not belong to any instance
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-12-13 07:25
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0020_pipelinemodel_priority"),
    ]

    operations = [
        migrations.AddField(
            model_name="pipelinemodel",
            name="queue",
            field=models.CharField(default="", max_length=512, verbose_name="流程使用的队列名"),
        ),
        migrations.AlterField(
            model_name="functionswitch", name="description", field=models.TextField(default="", verbose_name="功能描述"),
        ),
        migrations.AlterField(
            model_name="history",
            name="data",
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to="engine.HistoryData"),
        ),
        migrations.AlterField(
            model_name="nodecelerytask",
            name="celery_task_id",
            field=models.CharField(default="", max_length=40, verbose_name="celery 任务 ID"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="current_node_id",
            field=models.CharField(db_index=True, default="", max_length=32, verbose_name="当前推进到的节点的 ID"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="destination_id",
            field=models.CharField(default="", max_length=32, verbose_name="遇到该 ID 的节点就停止推进"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="parent_id",
            field=models.CharField(default="", max_length=32, verbose_name="父 process 的 ID"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="snapshot",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.SET_NULL, to="engine.ProcessSnapshot"
            ),
        ),
        migrations.AlterField(
            model_name="processcelerytask",
            name="celery_task_id",
            field=models.CharField(default="", max_length=40, verbose_name="celery 任务 ID"),
        ),
        migrations.AlterField(
            model_name="schedulecelerytask",
            name="celery_task_id",
            field=models.CharField(default="", max_length=40, verbose_name="celery 任务 ID"),
        ),
        migrations.AlterField(
            model_name="status", name="name", field=models.CharField(default="", max_length=64, verbose_name="节点名称"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from bamboo_engine.eri import Service, ExecutableEvent, Variable

from pipeline.component_framework.library import ComponentLibrary
from pipeline.core.flow import FlowNodeClsFactory
from pipeline.core.data.library import VariableLibrary

from pipeline.eri.imp.service import ServiceWrapper
from pipeline.eri.imp.executable_event import ExecutableEndEventWrapper
from pipeline.eri.imp.variable import VariableWrapper


class PipelinePluginManagerMixin:
    def get_service(self, code: str, version: str) -> Service:
        """
        根据代号与版本获取特定服务对象实例

        :param code: 服务唯一代号
        :type code: str
        :param version: 服务版本
        :type version: str
        :return: 服务对象实例
        :rtype: Service
        """
        comp_cls = ComponentLibrary.get_component_class(code, version)
        service = comp_cls.bound_service()
        return ServiceWrapper(service)

    def get_executable_end_event(self, code: str) -> ExecutableEvent:
        """
        根据代号获取特定可执行结束事件实例

        :param code: 可执行结束事件唯一代号
        :type code: str
        :return: 可执行结束事件实例
        :rtype: ExecutableEvent:
        """
        event_cls = FlowNodeClsFactory.get_node_cls(code)
        event = event_cls(id=None)
        return ExecutableEndEventWrapper(event)

    def get_compute_variable(self, code: str, key: str, value: Variable, additional_data: dict) -> Variable:
        """
        根据代号获取变量实例

        :param code: 唯一代号
        :type code: str
        :param key: 变量 key
        :type key: str
        :param value: 变量配置
        :type value: Any
        :param additional_data: 额外数据字典
        :type additional_data: dict
        :return: 变量实例
        :rtype: Variable
        """
        var_cls = VariableLibrary.get_var_class(code=code)
        return VariableWrapper(orginal_value=value, var_cls=var_cls, additional_data=additional_data)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pyparsing import (
    CaselessLiteral,
    Combine,
    Forward,
    Group,
    Keyword,
    Optional,
    ParseException,
    ParseResults,
    QuotedString,
    Suppress,
    Word,
    ZeroOrMore,
    alphanums,
    alphas,
    delimitedList,
    nums,
    oneOf,
)


class SubstituteVal(object):
    """
    Represents a token that will later be replaced by a context value.
    """

    def __init__(self, t):
        self._path = t[0]

    def get_val(self, context):
        if not context:
            # raise MissingVariableException(
            #     'context missing or empty'
            # )
            return self._path

        val = context

        try:
            for part in self._path.split(pathDelimiter):
                val = getattr(val, part) if hasattr(val, part) else val[part]

        except KeyError:
            raise MissingVariableException("no value supplied for {}".format(self._path))

        return val

    def __repr__(self):
        return "SubstituteVal(%s)" % self._path


# Grammar definition
pathDelimiter = "."
# match gcloud's variable
identifier = Combine(Optional("${") + Optional("_") + Word(alphas, alphanums + "_") + Optional("}"))
# identifier = Word(alphas, alphanums + "_")
propertyPath = delimitedList(identifier, pathDelimiter, combine=True)

and_ = Keyword("and", caseless=True)
or_ = Keyword("or", caseless=True)
in_ = Keyword("in", caseless=True)

lparen = Suppress("(")
rparen = Suppress(")")

binaryOp = oneOf("== != < > >= <= in notin issuperset notissuperset", caseless=True)("operator")

E = CaselessLiteral("E")
numberSign = Word("+-", exact=1)
realNumber = Combine(
    Optional(numberSign)
    + (Word(nums) + "." + Optional(Word(nums)) | ("." + Word(nums)))
    + Optional(E + Optional(numberSign) + Word(nums))
)

integer = Combine(Optional(numberSign) + Word(nums) + Optional(E + Optional("+") + Word(nums)))

# str_ = quotedString.addParseAction(removeQuotes)
str_ = QuotedString('"') | QuotedString("'")
bool_ = oneOf("true false", caseless=True)

simpleVals = (
    realNumber.setParseAction(lambda toks: float(toks[0]))
    | integer.setParseAction(lambda toks: int(toks[0]))
    | str_
    | bool_.setParseAction(lambda toks: toks[0] == "true")
    | propertyPath.setParseAction(lambda toks: SubstituteVal(toks))
)  # need to add support for alg expressions

propertyVal = simpleVals | (lparen + Group(delimitedList(simpleVals)) + rparen)

boolExpression = Forward()
boolCondition = Group(
    (Group(propertyVal)("lval") + binaryOp + Group(propertyVal)("rval")) | (lparen + boolExpression + rparen)
)
boolExpression << boolCondition + ZeroOrMore((and_ | or_) + boolExpression)


def double_equals_trans(lval, rval, operator):
    # double equals
    if operator in ["in", "notin"]:
        if isinstance(rval, list) and len(rval):
            transed_rval = []
            if isinstance(lval, int):
                for item in rval:
                    try:
                        transed_rval.append(int(item))
                    except Exception:
                        pass
            elif isinstance(lval, str):
                for item in rval:
                    try:
                        transed_rval.append(str(item))
                    except Exception:
                        pass
            rval += transed_rval

    elif operator in ["issuperset", "notissuperset"]:
        # avoid convert set('abc') to {a, b, c}, but keep {'abc'}
        if isinstance(lval, str):
            lval = [lval]
        if isinstance(rval, str):
            rval = [rval]

    else:
        try:
            if isinstance(lval, int):
                rval = int(rval)
            elif isinstance(rval, int):
                lval = int(lval)
            if isinstance(lval, str):
                rval = str(rval)
            elif isinstance(rval, str):
                lval = str(lval)
        except Exception:
            pass

    return lval, rval


class BoolRule(object):
    """
    Represents a boolean expression and provides a `test` method to evaluate
    the expression and determine its truthiness.

    :param query: A string containing the query to be evaluated
    :param lazy: If ``True``, parse the query the first time it's tested rather
                 than immediately. This can help with performance if you
                 instantiate a lot of rules and only end up evaluating a
                 small handful.
    """

    _compiled = False
    _tokens = None
    _query = None

    def __init__(self, query, lazy=False, strict=True):
        self._query = query
        self.strict = strict
        if not lazy:
            self._compile()

    def test(self, context=None):
        """
        Test the expression against the given context and return the result.

        :param context: A dict context to evaluate the expression against.
        :return: True if the expression succesfully evaluated against the
                 context, or False otherwise.
        """
        if self._is_match_all():
            return True

        self._compile()
        return self._test_tokens(self._tokens, context)

    def _is_match_all(self):
        return True if self._query == "*" else False

    def _compile(self):
        if not self._compiled:

            # special case match-all query
            if self._is_match_all():
                return

            try:
                self._tokens = boolExpression.parseString(self._query, parseAll=self.strict)
            except ParseException:
                raise

            self._compiled = True

    def _expand_val(self, val, context):
        if type(val) == list:
            val = [self._expand_val(v, context) for v in val]

        if isinstance(val, SubstituteVal):
            ret = val.get_val(context)
            return ret

        if isinstance(val, ParseResults):
            return [self._expand_val(x, context) for x in val.asList()]

        return val

    def _test_tokens(self, tokens, context):
        passed = False

        for token in tokens:

            if not isinstance(token, ParseResults):
                if token == "or" and passed:
                    return True
                elif token == "and" and not passed:
                    return False
                continue

            if not token.getName():
                passed = self._test_tokens(token, context)
                continue

            items = token.asDict()

            operator = items["operator"]
            lval = self._expand_val(items["lval"][0], context)
            rval = self._expand_val(items["rval"][0], context)
            lval, rval = double_equals_trans(lval, rval, operator)

            if operator in ("=", "==", "eq"):
                passed = lval == rval
            elif operator in ("!=", "ne"):
                passed = lval != rval
            elif operator in (">", "gt"):
                passed = lval > rval
            elif operator in (">=", "ge"):
                passed = lval >= rval
            elif operator in ("<", "lt"):
                passed = lval < rval
            elif operator in ("<=", "le"):
                passed = lval <= rval
            elif operator == "in":
                passed = lval in rval
            elif operator == "notin":
                passed = lval not in rval
            elif operator == "issuperset":
                passed = set(lval).issuperset(set(rval))
            elif operator == "notissuperset":
                passed = not set(lval).issuperset(set(rval))
            else:
                raise UnknownOperatorException("Unknown operator '{}'".format(operator))

        return passed


class MissingVariableException(Exception):
    """
    Raised when an expression contains a property path that's not supplied in
    the context.
    """

    pass


class UnknownOperatorException(Exception):
    """
    Raised when an expression uses an unknown operator.

    This should never be thrown since the operator won't be correctly parsed as
    a token by pyparsing, but it's useful to have this hanging around for when
    additional operators are being added.
    """

    pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy

from kombu import Exchange, Queue

from pipeline.celery.queues import ScalableQueues
from pipeline.constants import PIPELINE_MAX_PRIORITY

default_exchange = Exchange("default", type="direct")

# 设置时区
CELERY_TIMEZONE = "Asia/Shanghai"
# 启动时区设置
CELERY_ENABLE_UTC = False

# new priority queues
PUSH_DEFAULT_QUEUE_NAME = "pipeline_priority"
PUSH_DEFAULT_ROUTING_KEY = "pipeline_push_priority"

SCHEDULE_DEFAULT_QUEUE_NAME = "service_schedule_priority"
SCHEDULE_DEFAULT_ROUTING_KEY = "schedule_service_priority"

ADDITIONAL_DEFAULT_QUEUE_NAME = "pipeline_additional_task_priority"
ADDITIONAL_DEFAULT_ROUTING_KEY = "additional_task_priority"

STATISTICS_PRIORITY_QUEUE_NAME = "pipeline_statistics_priority"
STATISTICS_PRIORITY_ROUTING_KEY = "pipeline_statistics_priority"

SCALABLE_QUEUES_CONFIG = {
    PUSH_DEFAULT_QUEUE_NAME: {"name": PUSH_DEFAULT_QUEUE_NAME, "routing_key": PUSH_DEFAULT_ROUTING_KEY},
    SCHEDULE_DEFAULT_QUEUE_NAME: {"name": SCHEDULE_DEFAULT_QUEUE_NAME, "routing_key": SCHEDULE_DEFAULT_ROUTING_KEY},
}

PIPELINE_PRIORITY_ROUTING = {
    "queue": PUSH_DEFAULT_QUEUE_NAME,
    "routing_key": PUSH_DEFAULT_ROUTING_KEY,
}

PIPELINE_SCHEDULE_PRIORITY_ROUTING = {
    "queue": SCHEDULE_DEFAULT_QUEUE_NAME,
    "routing_key": SCHEDULE_DEFAULT_ROUTING_KEY,
}

PIPELINE_ADDITIONAL_PRIORITY_ROUTING = {
    "queue": ADDITIONAL_DEFAULT_QUEUE_NAME,
    "routing_key": ADDITIONAL_DEFAULT_ROUTING_KEY,
}

PIPELINE_STATISTICS_PRIORITY_ROUTING = {
    "queue": STATISTICS_PRIORITY_QUEUE_NAME,
    "routing_key": STATISTICS_PRIORITY_ROUTING_KEY,
}

CELERY_ROUTES = {
    # schedule
    "pipeline.engine.tasks.service_schedule": PIPELINE_SCHEDULE_PRIORITY_ROUTING,
    # pipeline
    "pipeline.engine.tasks.batch_wake_up": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.dispatch": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.process_wake_up": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.start": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.wake_from_schedule": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.wake_up": PIPELINE_PRIORITY_ROUTING,
    "pipeline.engine.tasks.process_unfreeze": PIPELINE_PRIORITY_ROUTING,
    # another
    "pipeline.log.tasks.clean_expired_log": PIPELINE_ADDITIONAL_PRIORITY_ROUTING,
    "pipeline.engine.tasks.node_timeout_check": PIPELINE_ADDITIONAL_PRIORITY_ROUTING,
    "pipeline.contrib.periodic_task.tasks.periodic_task_start": PIPELINE_ADDITIONAL_PRIORITY_ROUTING,
    "pipeline.contrib.periodic_task.tasks.bamboo_engine_periodic_task_start": PIPELINE_ADDITIONAL_PRIORITY_ROUTING,
    "pipeline.engine.tasks.heal_zombie_process": PIPELINE_ADDITIONAL_PRIORITY_ROUTING,
    # statistics
    "pipeline.contrib.statistics.tasks.pipeline_post_save_statistics_task": PIPELINE_STATISTICS_PRIORITY_ROUTING,
    "pipeline.contrib.statistics.tasks.pipeline_archive_statistics_task": PIPELINE_STATISTICS_PRIORITY_ROUTING,
}


class QueueResolver(object):
    def __init__(self, queue):
        self.queue = queue

    def default_setting_for(self, task, setting_key):
        if not isinstance(task, str):
            task = task.name

        return CELERY_ROUTES[task][setting_key]

    def resolve_task_routing_key(self, task):
        default_key = self.default_setting_for(task, "routing_key")
        default_queue = self.default_setting_for(task, "queue")

        if default_queue not in SCALABLE_QUEUES_CONFIG or not self.queue:
            return default_key

        return self.resolve_routing_key(default_key)

    def resolve_queue_name(self, default_name):
        if not self.queue:
            return default_name

        return "{}_{}".format(self.queue, default_name)

    def resolve_routing_key(self, default_key):
        if not self.queue:
            return default_key

        return "{}_{}".format(ScalableQueues.routing_key_for(self.queue), default_key)


USER_QUEUES = []

for name, queue in ScalableQueues.queues().items():
    queue_arguments = copy.copy(queue["queue_arguments"])
    queue_arguments["x-max-priority"] = PIPELINE_MAX_PRIORITY

    for config in SCALABLE_QUEUES_CONFIG.values():
        resolver = QueueResolver(name)
        USER_QUEUES.append(
            Queue(
                resolver.resolve_queue_name(config["name"]),
                default_exchange,
                routing_key=resolver.resolve_routing_key(config["routing_key"]),
                queue_arguments=queue_arguments,
            )
        )

CELERY_QUEUES = [
    # user queues
    *USER_QUEUES,  # noqa
    # keep old queue to process message left in broker, remove on next version
    Queue("default", default_exchange, routing_key="default"),
    Queue("pipeline", default_exchange, routing_key="pipeline_push"),
    Queue("service_schedule", default_exchange, routing_key="schedule_service"),
    Queue("pipeline_additional_task", default_exchange, routing_key="additional_task"),
    # priority queues
    Queue(
        PUSH_DEFAULT_QUEUE_NAME,
        default_exchange,
        routing_key=PUSH_DEFAULT_ROUTING_KEY,
        queue_arguments={"x-max-priority": PIPELINE_MAX_PRIORITY},
    ),
    Queue(
        SCHEDULE_DEFAULT_QUEUE_NAME,
        default_exchange,
        routing_key=SCHEDULE_DEFAULT_ROUTING_KEY,
        queue_arguments={"x-max-priority": PIPELINE_MAX_PRIORITY},
    ),
    Queue(
        ADDITIONAL_DEFAULT_QUEUE_NAME,
        default_exchange,
        routing_key=ADDITIONAL_DEFAULT_ROUTING_KEY,
        queue_arguments={"x-max-priority": PIPELINE_MAX_PRIORITY},
    ),
    Queue(
        STATISTICS_PRIORITY_QUEUE_NAME,
        default_exchange,
        routing_key=STATISTICS_PRIORITY_ROUTING_KEY,
        queue_arguments={"x-max-priority": PIPELINE_MAX_PRIORITY},
    ),
]

CELERY_DEFAULT_QUEUE = "default"
CELERY_DEFAULT_EXCHANGE = "default"
CELERY_DEFAULT_ROUTING_KEY = "default"

CELERYBEAT_SCHEDULER = "django_celery_beat.schedulers.DatabaseScheduler"

CELERY_ACCEPT_CONTENT = ["json", "pickle", "msgpack", "yaml"]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import sys
from abc import abstractmethod
from copy import deepcopy

from django.db import IntegrityError, models
from django.utils.translation import ugettext_lazy as _

from pipeline.component_framework.library import ComponentLibrary
from pipeline.contrib.external_plugins import exceptions
from pipeline.contrib.external_plugins.models.fields import JSONTextField

GIT = "git"
S3 = "s3"
FILE_SYSTEM = "fs"
logger = logging.getLogger("root")
source_cls_factory = {}


def package_source(cls):
    source_cls_factory[cls.type()] = cls
    return cls


class SourceManager(models.Manager):
    def create_source(self, name, packages, from_config, **kwargs):
        create_kwargs = deepcopy(kwargs)
        create_kwargs.update({"name": name, "packages": packages, "from_config": from_config})
        return self.create(**create_kwargs)

    def remove_source(self, source_id):
        source = self.get(id=source_id)

        if source.from_config:
            raise exceptions.InvalidOperationException("Can not remove source create from config")

        source.delete()

    def update_source_from_config(self, configs):

        sources_from_config = self.filter(from_config=True).all()
        existing_source_names = {source.name for source in sources_from_config}
        source_name_in_config = {config["name"] for config in configs}

        invalid_source_names = existing_source_names - source_name_in_config

        # remove invalid source
        self.filter(name__in=invalid_source_names).delete()

        # update and create source
        for config in configs:
            defaults = deepcopy(config["details"])
            defaults["packages"] = config["packages"]

            try:
                self.update_or_create(name=config["name"], from_config=True, defaults=defaults)
            except IntegrityError:
                raise exceptions.InvalidOperationException(
                    'There is a external source named "{source_name}" but not create from config, '
                    "can not do source update operation".format(source_name=config["name"])
                )


class ExternalPackageSource(models.Model):
    name = models.CharField(_("包源名"), max_length=128, unique=True)
    from_config = models.BooleanField(_("是否是从配置文件中读取的"), default=False)
    packages = JSONTextField(_("模块配置"))

    objects = SourceManager()

    class Meta:
        abstract = True

    @staticmethod
    @abstractmethod
    def type():
        raise NotImplementedError()

    @abstractmethod
    def importer(self):
        raise NotImplementedError()

    @abstractmethod
    def details(self):
        raise NotImplementedError()

    @property
    def imported_plugins(self):
        plugins = []
        try:
            importer = self.importer()
        except ValueError as e:
            logger.exception("ExternalPackageSource[name={}] call importer error: {}".format(self.name, e))
            return plugins
        for component in ComponentLibrary.component_list():
            component_importer = getattr(sys.modules[component.__module__], "__loader__", None)
            if isinstance(component_importer, type(importer)) and component_importer.name == self.name:
                plugins.append(
                    {
                        "code": component.code,
                        "name": component.name,
                        "group_name": component.group_name,
                        "class_name": component.__name__,
                        "module": component.__module__,
                    }
                )
        return plugins

    @property
    def modules(self):
        modules = []

        for package_info in list(self.packages.values()):
            modules.extend(package_info["modules"])

        return modules

    @staticmethod
    def update_package_source_from_config(source_configs):
        classified_config = {source_type: [] for source_type in list(source_cls_factory.keys())}

        for config in deepcopy(source_configs):
            classified_config.setdefault(config.pop("type"), []).append(config)

        for source_type, configs in list(classified_config.items()):
            try:
                source_model_cls = source_cls_factory[source_type]
            except KeyError:
                raise KeyError("Unsupported external source type: %s" % source_type)
            source_model_cls.objects.update_source_from_config(configs=configs)

    @staticmethod
    def package_source_types():
        return list(source_cls_factory.keys())



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from os import path

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
version = __import__("pipeline").__version__

setup(
    name="bamboo-pipeline",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,
    description="pipeline",  # noqa
    long_description="runtime for bamboo-engine base on Django and Celery",
    # The project's main homepage.
    url="https://github.com/Tencent/bk-sops/tree/sdk/sdk/bamboo-pipeline",
    # Author details
    author="Blueking",
    author_email="howellliang@tencent.com",
    include_package_data=True,
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        # Base
        "Django>=2.2.6,<3.0",
        "requests>=2.22.0,<=2.23.0",
        "celery>=4.4.0,<5.0",
        "django-celery-beat==2.0.0",
        "django-celery-results==1.2.1",
        "Mako>=1.0.6,<2.0",
        "pytz==2019.3",
        # pipeline
        "jsonschema==2.5.1",
        "ujson==1.35",
        "pyparsing==2.2.0",
        "redis==3.2.0",
        "redis-py-cluster==2.1.0",
        "django-timezone-field==4.0",
        "boto3==1.9.130",
        "Werkzeug==1.0.1",
    ],
    zip_safe=False,
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     "console_scripts": ["bk-admin=blueapps.contrib.bk_commands:bk_admin"],
    # },
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-11-23 07:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0026_auto_20201028_1049"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pipelineinstance",
            name="instance_id",
            field=models.CharField(db_index=True, max_length=32, unique=True, verbose_name="实例ID"),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-24 03:08
from __future__ import unicode_literals

from django.db import migrations

import pipeline.engine.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0023_status_state_refresh_at"),
    ]

    operations = [
        migrations.AlterField(
            model_name="data",
            name="ex_data",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="异常数据"),
        ),
        migrations.AlterField(
            model_name="data",
            name="inputs",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="输入数据"),
        ),
        migrations.AlterField(
            model_name="data",
            name="outputs",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="输出数据"),
        ),
        migrations.AlterField(
            model_name="historydata",
            name="ex_data",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="异常数据"),
        ),
        migrations.AlterField(
            model_name="historydata",
            name="inputs",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="输入数据"),
        ),
        migrations.AlterField(
            model_name="historydata",
            name="outputs",
            field=pipeline.engine.models.fields.IOField(default=None, verbose_name="输出数据"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import importlib
import logging

from django.db.utils import ProgrammingError

from pipeline.component_framework.constants import LEGACY_PLUGINS_VERSION
from pipeline.component_framework.library import ComponentLibrary
from pipeline.component_framework.models import ComponentModel
from pipeline.component_framework import context
from pipeline.core.flow.activity import Service

logger = logging.getLogger(__name__)


class ComponentMeta(type):
    def __new__(cls, name, bases, attrs):
        super_new = super(ComponentMeta, cls).__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, ComponentMeta)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Create the class
        module_name = attrs.pop("__module__")
        new_class = super_new(cls, name, bases, {"__module__": module_name})
        module = importlib.import_module(new_class.__module__)

        # Add all attributes to the class
        attrs.setdefault("desc", "")
        for obj_name, obj in list(attrs.items()):
            setattr(new_class, obj_name, obj)

        # check
        if not getattr(new_class, "name", None):
            raise ValueError("component %s name can't be empty" % new_class.__name__)

        if not getattr(new_class, "code", None):
            raise ValueError("component %s code can't be empty" % new_class.__name__)

        if not getattr(new_class, "bound_service", None) or not issubclass(new_class.bound_service, Service):
            raise ValueError("component %s service can't be empty and must be subclass of Service" % new_class.__name__)

        if not getattr(new_class, "form", None):
            setattr(new_class, "form", None)

        if not getattr(new_class, "output_form", None):
            setattr(new_class, "output_form", None)

        if not getattr(new_class, "version", None):
            setattr(new_class, "version", LEGACY_PLUGINS_VERSION)

        # category/group name
        group_name = getattr(module, "__group_name__", new_class.__module__.split(".")[-1].title())
        setattr(new_class, "group_name", group_name)
        new_name = "{}-{}".format(group_name, new_class.name)

        # category/group name
        group_icon = getattr(module, "__group_icon__", "")
        setattr(new_class, "group_icon", group_icon)

        if not getattr(module, "__register_ignore__", False):
            ComponentLibrary.register_component(
                component_code=new_class.code, version=new_class.version, component_cls=new_class
            )

            if context.skip_update_comp_models():
                return new_class

            try:
                print("update {} component model".format(new_class.code))
                ComponentModel.objects.update_or_create(
                    code=new_class.code, version=new_class.version, defaults={"name": new_name, "status": __debug__}
                )
            except Exception as e:
                if not isinstance(e, ProgrammingError):
                    logging.exception(e)

        return new_class
